#!/usr/bin/env python

import os
import torch
import torchreid

import numpy as np

from typing import List

from containers import Tracker

class ReIDManager(object):
    def __init__(self, model_path : str, model_name : str = "resnet50", threshold=0.75, lower_threshold=0.7, img_size=(256,128)) -> None:

        if not os.path.exists(model_path):
            raise FileNotFoundError(f"The model path: {model_path} doenst exist")
        self.__trackers : List[Tracker] = []
        self.__model_path : str = model_path
        self.__model_name = model_name
        self.__threshold = threshold
        self.__lower_threshold = lower_threshold
        self.__trackers_counter = 0
        self.__img_size = img_size
        self.__loadExtractor()

    def __loadExtractor(self):
        self.__extractor = torchreid.utils.FeatureExtractor(
            self.__model_name,
            self.__model_path,
            image_size=self.__img_size
        )

    def extract_id(self, track_id : int, img_patch : np.ndarray) -> int:
        feature = self.__extractor(img_patch)
        for tracker in self.__trackers:
            if track_id == tracker.track_id:
                if self.__lower_threshold < tracker.getDistance(feature) < self.__threshold:
                    tracker.addFeature(feature)
                return tracker.id
        distances = []
        for tracker in self.__trackers:
            distances.append(tracker.getDistance(feature))
        max_dist = 0
        if len(distances) != 0:
            max_dist = max(distances)
        if max_dist > self.__threshold:
            self.__trackers[distances.index(max_dist)].track_id = track_id
            return self.__trackers[distances.index(max_dist)].id
        else:
            self.__trackers.append(Tracker(self.__trackers_counter))
            self.__trackers[-1].addFeature(feature)
            self.__trackers[-1].track_id = track_id
            self.__trackers_counter += 1 
            return self.__trackers[-1].id





#!/usr/bin/env python

import torch

from typing import List

class Tracker(object):
    def __init__(self, id : int) -> None:
        self.__id = id
        self.__track_id = id
        self.__features : List[torch.tensor] = []

    @property
    def id(self) -> int:
        return self.__id
    
    @property
    def track_id(self) -> int:
        return self.__track_id

    @track_id.setter
    def track_id(self, id) -> None:
        self.__track_id = id

    def addFeature(self, feature : torch.tensor) -> None:
        self.__features.append(feature)
        return
    
    def getDistance(self, feature : torch.tensor) -> float:
        distances = []
        for feature_i in self.__features:
            distances.append(torch.nn.functional.cosine_similarity(feature, feature_i).item())
        return max(distances)


#!/usr/bin/env python
import cv2 as cv
import numpy as np

from ultralytics import YOLO
from ultralytics.engine.results import Results
from ReIDManager import ReIDManager

MODEL_PATH = "./models/model.pt"
YOLO_MODEL = "./yolov8n-pose.pt"
INPUT_FILE = "./sample.mp4"
OUTPUT_FILE = "./output.mp4"



def main() -> None:
    cap = cv.VideoCapture(INPUT_FILE)
    fps = cap.get(cv.CAP_PROP_FPS)
    height = int(cap.get(cv.CAP_PROP_FRAME_HEIGHT))
    width = int(cap.get(cv.CAP_PROP_FRAME_WIDTH))
    total_frames = int(cap.get(cv.CAP_PROP_FRAME_COUNT))

    manager = ReIDManager(MODEL_PATH)

    print(f"Video Info:\n\tResolution: {width}x{height}\n\tFPS: {fps:.02f}\n\tTotal Frames: {total_frames-1}")

    writer = cv.VideoWriter(OUTPUT_FILE,cv.VideoWriter_fourcc(*'mp4v'), fps,(width,height))

    yolo_model = YOLO(YOLO_MODEL)

    for i in range(total_frames):
        print(f"\rFrames: {i:04d}/{total_frames:04d}", end="")
        ret, frame = cap.read()
        if not ret:
            break
        results = yolo_model.track(frame,persist=True,verbose=False)[0]

        if results.boxes.is_track:         

            result : Results
            for result in results:
                for (x1,y1,x2,y2), id, score in zip(result.boxes.xyxy.int().cpu().numpy(), result.boxes.id.int().cpu().tolist(), result.boxes.conf.cpu().tolist()):
                    if score < 0.75:
                        continue
                    id = manager.extract_id(id, frame[y1:y2,x1:x2])
                    frame = cv.rectangle(frame,(int(x1),int(y1)),(int(x2),int(y2)),(255,0,0),3)
                    cv.putText(frame,f"ID: {id}", (x1,y1),cv.FONT_HERSHEY_PLAIN,4, (255,0,0),4)
        writer.write(frame)

    print("\nDone")
    writer.release()
if __name__ == "__main__":
    main()